@@ -1,8 +1,10 @@
 """Main module."""
+from math import tau
 import numpy as np
 import cupy as cp
 from timeit import default_timer as timer
 import matplotlib.pyplot as plt
+from taufactor.metrics import surface_area
 
 class Solver:
     """
@@ -35,7 +37,6 @@ class Solver:
 
         # init conc
         self.conc = self.init_conc(img)
-
         # create nn map
         self.nn = self.init_nn(img)
 
@@ -137,10 +138,7 @@ class Solver:
         self.D_mean = self.D_0
         self.tau=self.VF/D_rel if D_rel != 0 else cp.inf
         self.D_eff=self.D_mean*D_rel
-        if verbose:
-            print('converged to:', self.tau,
-                  'after: ', self.iter, 'iterations in: ', np.around(timer() - start, 4),
-                  'seconds at a rate of', np.around((timer() - start)/self.iter, 4), 's/iter')
+        self.end_simulation(iter_limit, verbose, start)
         return self.tau
 
     def check_convergence(self, lt, lb, verbose, conv_crit, start, iter_limit):
@@ -172,7 +170,6 @@ class Solver:
                 self.nn = cp.array(self.nn, dtype=cp.double)
                 self.precision = cp.double
             else:
-                print('Did not converge in the iteration limit')
                 return True, ((lt + lb) * self.L_A / abs(self.top_bc - self.bot_bc)).get()
 
         return False, False
@@ -217,6 +214,17 @@ class Solver:
         plt.imshow(flux[:, :, lay])
         return flux
 
+    def end_simulation(self, iter_limit, verbose, start):
+        converged = 'converged to'
+        if self.iter >=iter_limit -1:
+            print('Warning: not converged')
+            converged = 'unconverged value of tau'
+        
+        if verbose:
+            print(f'{converged}: {self.tau} \
+                  after: {self.iter} iterations in: {np.around(timer() - start, 4)}  \
+                  seconds at a rate of {np.around((timer() - start)/self.iter, 4)} s/iter')
+
 class PeriodicSolver(Solver):
     """
     Periodic Solver (works for non-periodic structures, but has higher RAM requirements)
@@ -280,10 +288,7 @@ class PeriodicSolver(Solver):
         self.D_mean=D_0
         self.tau = self.VF/D_rel if D_rel !=0 else cp.inf
         self.D_eff=D_0*D_rel
-        if verbose:
-            print('converged to:', self.tau,
-                  'after: ', self.iter, 'iterations in: ', np.around(timer() - start, 4),
-                  'seconds at a rate of', np.around((timer() - start)/self.iter, 4), 's/iter')
+        self.end_simulation(iter_limit, verbose, start)
         return self.tau
 
     def check_vertical_flux(self, conv_crit):
@@ -428,12 +433,8 @@ class MultiPhaseSolver(Solver):
         else:
             self.D_mean=0
         self.tau = self.D_mean/self.D_eff if self.D_eff != 0 else cp.inf
-        if verbose:
-            print('converged to:', self.tau,
-                  'after: ', self.iter, 'iterations in: ', np.around(timer() - start, 4),
-                  'seconds at a rate of', np.around((timer() - start)/self.iter, 4), 's/iter')
+        self.end_simulation(iter_limit, verbose, start)
         return self.tau
-
     def check_convergence(self, verbose, conv_crit, start, iter_limit):
         # print progress
         if self.iter % 100 == 0:
@@ -455,7 +456,6 @@ class MultiPhaseSolver(Solver):
                 self.nn = cp.array(self.nn, dtype=cp.double)
                 self.precision = cp.double
             else:
-                print('Did not converge in the iteration limit')
                 return True, flux.get()
         return False, False
 
@@ -468,11 +468,254 @@ class MultiPhaseSolver(Solver):
             return 0, cp.array([0], dtype=self.precision)
         return err, fl.mean()
 
-img = np.ones([64,64,64])
-img[32:, :] = 1
-# img[:, 32:, :] = 2
-# img[32:, 32:, :] = 3
-s = Solver(img)
-# tau = s.solve()
-# s = PeriodicSolver(img)
-s.solve()
+
+class ElectrodeSolver():
+    """
+    Electrode Solver - solves the electrode tortuosity factor system (migration and capacitive current between current collector and solid/electrolyte interface)
+    Once solve method is called, tau, D_eff and D_rel are available as attributes.
+    """
+    
+    def __init__(self, img, precision=cp.double, omega=1e-6):
+
+        img = np.expand_dims(img, 0)
+        self.cpu_img = img
+        self.precision = precision
+        
+        # Define omega, res and c_DL
+        self.omega = omega
+        self.res = 1
+        self.c_DL = 1
+        if len(img.shape)==4:
+            self.A_CC = img.shape[2]*img.shape[3]
+        else:
+            self.A_CC = img.shape[2]
+        self.k_0 = 1
+        
+        # VF calc
+        self.VF = np.mean(img)
+
+        # save original image in cuda
+        img = cp.array(img, dtype=self.precision)
+        self.img = img
+        # init phi
+        
+        self.phi = self.init_phi(img)
+        
+        self.phase_map = self.pad(img,[1,0])
+
+        # create prefactor map
+        self.prefactor = self.init_prefactor(img)
+
+        
+
+        #checkerboarding
+        self.w = 2 - cp.pi / (1.5 * img.shape[1])
+        # self.w = 1.8
+        # self.w = 0.01
+        self.cb = self.init_cb(img)
+
+        # solving params
+        self.converged = False
+        self.semiconverged = False
+        self.iter=0
+
+        # Results
+        self.tau_e=0
+        self.D_eff=None
+        self.D_mean=None
+    
+    def pad(self, img, vals=[0] * 6):
+        while len(vals) < 6:
+            vals.append(0)
+        if len(img.shape)==4:
+            to_pad = [1]*8
+            to_pad[:2] = (0, 0)
+        elif len(img.shape)==3:
+            to_pad = [1]*6
+            to_pad[:2] = (0, 0)
+
+        img = cp.pad(img, to_pad, 'constant')
+        img[:, 0], img[:, -1] = vals[:2]
+        img[:, :, 0], img[:, :, -1] = vals[2:4]
+
+        if len(img.shape)==4:
+            img[:, :, :, 0], img[:, :, :, -1] = vals[4:]
+        return img
+    
+    def crop(self, img, c = 1):
+        if len(img.shape)==4:
+            return img[:, c:-c, c:-c, c:-c]
+        elif len(img.shape)==3:
+            return img[:, c:-c, c:-c]
+    
+    def init_phi(self, img):
+        """
+        Initialise phi field as zeros
+
+        :param img: input image, with 1s conductive and 0s non-conductive
+        :type img: cp.array
+        :return: phi
+        :rtype: cp.array
+        """
+        phi = cp.zeros_like(img, dtype=self.precision)+0j
+        phi = self.pad(phi, [1, 0])
+        return phi
+    
+    def init_cb(self, img):
+
+        if len(img.shape)==4:
+            bs, x, y, z = img.shape
+            cb = np.zeros([x, y, z])
+            a, b, c = np.meshgrid(range(x), range(y), range(z), indexing='ij')
+            cb[(a + b + c) % 2 == 0] = 1*self.w
+            return [cp.roll(cp.array(cb), sh, 0) for sh in [0, 1]]
+
+        elif len(img.shape)==3:
+            bs, x, y = img.shape
+            cb = np.zeros([x, y])
+            a, b = np.meshgrid(range(x), range(y), indexing='ij')
+            cb[(a + b) % 2 == 0] = 1*self.w
+            cb = [cp.roll(cp.array(cb), sh, 0) for sh in [0, 1]]
+            cb[1][0] = cb[1][2]
+            return cb
+
+
+    def init_prefactor(self, img):
+        """
+        Initialise prefactors -> (nn_cond+2j*omega*res*c(dims-nn_cond))**-1
+
+        :param img: input image, with 1s conductive and 0s non-conductive
+        :type img: cp.array
+        :return: prefactor
+        :rtype: cp.array
+        """
+        dims = (len(img.shape)-1)*2
+        # find number of conducting nearest neighbours
+        img2 = self.pad(img, [1,0])
+        nn_cond = cp.zeros_like(img2, dtype=self.precision)
+        # iterate through shifts in the spatial dimensions
+        for dim in range(1, len(img.shape)):
+            for dr in [1, -1]:
+                nn_cond += cp.roll(img2, dr, dim)
+        # remove the paddings
+        nn_cond = self.crop(nn_cond, 1)
+        self.nn = nn_cond
+        orc = self.omega*self.res*self.c_DL
+        nn_solid = dims - nn_cond
+        omegapf = (orc**2 + 1j*orc)/(orc**2+1)
+        prefactor = (nn_cond + 2*nn_solid*omegapf)**-1
+        # prefactor = (nn_cond+2j*self.omega*self.res*self.c_DL*(dims-nn_cond))**-1
+        prefactor[prefactor==cp.inf] = 0
+        prefactor[img==0] = 0
+        return prefactor
+    
+    def sum_neighbours(self):
+        i=0
+        for dim in range(1, len(self.phi.shape)):
+            for dr in [1, -1]:
+                if i==0:
+                    out = cp.roll(self.phi, dr, dim)
+                else:
+                    out += cp.roll(self.phi, dr, dim)
+                i+=1
+
+        out = self.crop(out, 1)
+        return out
+    
+    def check_convergence(self):
+        
+        if len(self.tau_es)  < 1000:
+            return False
+        loss = np.std(np.array(self.tau_es[-100:]))
+        # print(len(self.tau_es),self.tau_es[-1], loss)
+        if self.verbose=='per_iter':
+            print(f'(iter {self.iter} loss {loss}, taue {self.tau_es[-1]}')
+        if loss < self.conv_crit:
+            if self.semiconverged:
+                if self.tau_es[-1] > 1e-5:
+                    if abs(self.semiconverged - self.tau_es[-1]) < self.conv_crit_2:
+                        self.tau_e = self.tau_es[-1]
+                        self.end_simulation()
+                        return True
+                else:
+                    self.phi = self.init_phi(self.img)
+            self.semiconverged = self.tau_es[-1]
+            self.omega *= 0.1
+            print(f'Semi-converged to {self.semiconverged}. Reducing omega to {self.omega} to check convergence')
+            
+            self.iter = 0
+            self.prefactor = self.init_prefactor(self.img)
+            self.solve(iter_limit=self.iter_limit, verbose=self.verbose, conv_crit=self.conv_crit)
+            return True
+        if self.iter_limit == self.iter:
+            print('Iteration limit reached. Increase the iteration limit or try starting from a smaller omega')
+            
+            return True
+        return False
+
+    
+    def tau_e_from_phi(self):
+        #  calculate total current on bottom boundary
+        n = self.phase_map[0,1].sum()
+        z = self.res / (n-self.phi[0, 1].sum())
+        self.z = z
+        r_ion = z.real*3
+        tau_e = self.VF * r_ion * self.k_0 * self.A_CC / self.phi.shape[1]
+
+        return tau_e.get()
+
+
+
+    def solve(self, iter_limit=100000, verbose=True, conv_crit=1e-5, conv_crit_2=1e-3):
+        """
+        run a solve simulation
+
+        :param iter_limit: max iterations before aborting, will attempt double for the same no. iterations
+        if initialised as singles
+        :param verbose: Whether to print tau. Can be set to 'per_iter' for more feedback
+        :param conv_crit: convergence criteria - running standard deviation of tau_e
+        :param conv_crit_2: convergence criteria - maximum difference between tau_e in consecutive omega solves
+        :return: tau
+        """
+        self.conv_crit = conv_crit
+        self.conv_crit_2 = conv_crit_2
+        
+        self.iter_limit = iter_limit
+        self.verbose = verbose
+        dim = len(self.phi.shape)
+        self.start = timer()
+        self.frames = []
+        self.loss = []
+        self.tau_es = []
+
+        while not self.converged:
+            out = self.sum_neighbours()
+            out *= self.prefactor*self.crop(self.phase_map)
+            out[self.prefactor==-1] = 0
+            self.tau_es.append(self.tau_e_from_phi())
+            if self.iter % 100 == 0:
+                self.converged = self.check_convergence()
+            out -= self.crop(self.phi, 1)
+            out *= self.cb[self.iter%2]
+
+            if dim==4:
+                self.phi[:, 1:-1, 1:-1, 1:-1] += out
+            elif dim==3:
+                self.phi[:, 1:-1, 1:-1] += out
+
+            self.iter += 1
+        # self.tau_e = self.tau_es[-1]
+        # self.end_simulation(iter_limit, verbose, start)
+    
+    def end_simulation(self, ):
+        if self.iter==self.iter_limit -1:
+            print('Warning: not converged')
+            converged = 'unconverged value of tau'
+        converged = 'converged to'
+        if self.verbose:
+            print(f'{converged}: {self.tau_e} after: {self.iter} iterations in: {np.around(timer() - self.start, 4)} seconds at a rate of {np.around((timer() - self.start)/self.iter, 4)} s/iter')
+
+
+
+
+
